#include <stdlib.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <packet.h>
#include <cleanup.h>
#include <capdev.h>
#include <capture.h>
#include <signature.h>
#include <decode.h>
#include <plugin.h>
#include <args.h>
#include <pcap.h>

#define READ_TIMEOUT 500

struct fpcap_priv {
	pcap_t		*pcap_desc;
	struct packet	pkt;
	struct proto	*proto;
};

int cb_if(struct arg *, void *);
struct arg pl_args[]={
	{"if",ARGTYPE_STRING,cb_if},
	{NULL,ARGTYPE_NOP,NULL}
};

proc_args_parse args_parse;
proc_decode_subproto decode_subproto;
proc_serial_number serial_number;

int cb_if(struct arg *a, void *priv)
{
	struct fpcap_priv *p=priv;
	
	unsigned int lnk;
	char ebuf[PCAP_ERRBUF_SIZE];

	if ( !p ) return 0;

	if ( p->pcap_desc ) {
		printf("pcap: Can't specify two interfaces!\n");
		return 0;
	}

	if ( !(p->pcap_desc=pcap_open_live(a->val.v_str, 4096,
		0, /* Not promiscous */
		READ_TIMEOUT, ebuf)) ) {
		printf("pcap: %s\n", ebuf);
		return 0;
	}

	lnk=pcap_datalink(p->pcap_desc);
	
	if ( !(p->proto=decode_subproto("__pcap_dlt",lnk)) ) {
		printf("pcap: %s: Can't support protocol 0x%x\n",
			a->val.v_str, lnk);
		pcap_close(p->pcap_desc);
		return 0;
	}

	return 1;
}

void *fpcap_init(char *args)
{
	struct fpcap_priv *p=NULL;
	
	if ( !(p=calloc(1, sizeof(*p))) )
		return NULL;

	switch( args_parse(pl_args, args, p) ) {
	case -1:
		printf("pcap: parse error: %s\n", args);
	case 0: /* fall through */
		free(p);
		p=NULL;
		return 0;
	default:
		p->pkt.flags=FP_CLONE|FP_LIVE|FP_PROMISC;
		return p;
	}
}

void fpcap_end(void *priv)
{
	struct fpcap_priv *p=priv;

	if ( p && p->pcap_desc ) pcap_close(p->pcap_desc);
	if ( p ) free(p);
}

void lpf_callback(u_char *user, struct pcap_pkthdr *header, u_char *data)
{
	struct fpcap_priv *p;

	if ( !(p=(struct fpcap_priv *)user) ) return;

	/* Pointers */
	p->pkt.base=data;
	p->pkt.end=data;
	p->pkt.end+=header->caplen;

	/* length */	
	p->pkt.caplen=header->caplen;
	p->pkt.len=header->len;
	
	/* Timestamp */
	p->pkt.time.tv_sec=header->ts.tv_sec;
	p->pkt.time.tv_usec=header->ts.tv_usec;

	/* first layer */
	p->pkt.layer[0].proto=p->proto;
	p->pkt.layer[0].h.raw=p->pkt.base;
	p->pkt.layer[0].flags=0;
	p->pkt.layer[0].session=NULL;
	p->pkt.llen=1;
	p->pkt.bottom=0;
	p->pkt.top=0;

	serial_number(&p->pkt.serial);
	p->proto->decode(&p->pkt);
}

void fpcap_go(void *priv, struct cap_s *c)
{
	struct fpcap_priv *p=priv;
	p->pkt.capture=c;

	while(c->state==CAP_STATE_CAPTURE) {
		if ( pcap_dispatch(p->pcap_desc, 1,
			(pcap_handler)lpf_callback, priv)<0 ) {
			c->state=CAP_STATE_STOP;
			break;
		}
	}
}

int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	int ok=0;
	proc_capdev_add capdev_add;

	/* validate input */
	plugin_check(in, out);

	/* tell firestorm who we are */
	PLUGIN_ID("capture.pcap", "Live libpcap capture");
	PLUGIN_VERSION(1, 0);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@scaramanga.co.uk");
	PLUGIN_LICENSE("GPL");

	if ( !(capdev_add=in->import("capdev.add")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(decode_subproto=in->import("decode.subproto")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(serial_number=in->import("serial_number")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(args_parse=in->import("args.parse")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	ok+=capdev_add("pcap", fpcap_init, fpcap_end, fpcap_go);
	
	return (ok) ? PLUGIN_ERR_OK : PLUGIN_ERR_FAIL ;
}

int _PLUGIN_UNLOAD (int code) {
	return PLUGIN_ERR_OK;
}
