#ifndef __PLUGIN_TCPIP__
#define __PLUGIN_TCPIP__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <packet.h>
#include <signature.h>
#include <matcher.h>
#include <template.h>
#include <target.h>
#include <alert.h>
#include <plugin.h>
#include <decode.h>
#include <capdev.h>
#include <preproc.h>
#include <args.h>

/* Imported functions */
extern unsigned short ipfmask;
extern proc_preproc_activate preproc_activate;
extern proc_preproc_dispatch preproc_dispatch;
extern proc_serial_number serial_number;
extern proc_matcher_find matcher_find;
extern proc_args_parse args_parse;
extern proc_alert alert;

extern unsigned int use_ipfrag;
extern unsigned int tcp_stateful;

int ipfrag_init(char *);
int tcpstream_init(char *);

int ipfrag_process(struct packet *, unsigned int);
int tcpstream_process(struct packet *, unsigned int);
void icmperr_process(struct packet *, unsigned int);

void ipfrag_free(void);
void tcpstream_free(void);

void ipq_kill(struct ipq *);
void tcp_free(struct tcp_session *);

/* Protocols */
extern struct proto ipv4_p;
extern struct proto icmp_p;
extern struct proto tcp_p;
extern struct proto udp_p;

/* Protocol requests */
extern struct proto_req ipv4_r[];
extern struct proto_req icmp_r[];
extern struct proto_req tcp_r[];
extern struct proto_req udp_r[];

/* Decode functions */
void ipv4_decode(struct packet *);
void icmp_decode(struct packet *);
void tcp_decode(struct packet *);
void udp_decode(struct packet *);

/* Print functions */
int ipv4_dprint(struct layer*, char *, int);
int icmp_dprint(struct layer*, char *, int);
int tcp_dprint(struct layer*, char *, int);
int udp_dprint(struct layer*, char *, int);

/* Generators */
extern struct generator ip_gen;
extern struct generator dip_gen;
extern struct generator tcp_gen;
extern struct generator udp_gen;
extern struct generator icmp_gen;
extern struct generator ipfrag_gen;

/* Match functions */
void ip_match(struct packet *, unsigned int);
void tcp_match(struct packet *, unsigned int);
void udp_match(struct packet *, unsigned int);
void icmp_match(struct packet *, unsigned int);

/* Rule commital */
int ip_commit(struct rule *);
int tcp_commit(struct rule *);
int udp_commit(struct rule *);
int icmp_commit(struct rule *);

/* Support functions */
int parse_ports(char *str, u_int16_t *, u_int16_t*);
int parse_ipaddr(char *, u_int32_t *, u_int32_t *);
int parse_ipproto(char *, u_int8_t *);

/* No more! */
#define NEGATE_DADDR	(1<<0)
#define NEGATE_SADDR	(1<<1)
#define NEGATE_PROTO	(1<<2)
#define NEGATE_SPORT	(1<<3)
#define NEGATE_ITYPE	NEGATE_SPORT
#define NEGATE_DPORT	(1<<4)
#define NEGATE_ICODE	NEGATE_DPORT
#define CHECK_PROTO	(1<<5)
#define CHECK_ITYPE	(1<<6)
#define CHECK_ICODE	(1<<7)

/* Stored in NETWORK byte order */
struct ipm_head {
	struct ipm_head	*next;
	u_int32_t	saddr;
	u_int32_t	smask;
	u_int32_t	daddr;
	u_int32_t	dmask;
	u_int8_t	proto;
	u_int8_t	flags;
	struct ip_rule	*rules;
	struct ip_rule	*rules_tail;
};

/* For ICMP */
struct icmp_head {
	struct icmp_head *next;
	u_int32_t	saddr;
	u_int32_t	smask;
	u_int32_t	daddr;
	u_int32_t	dmask;
	u_int8_t	itype;
	u_int8_t	icode;
	u_int8_t	flags;
	struct ip_rule	*rules;
	struct ip_rule	*rules_tail;
};

/* For protocols with ports */
struct tcp_head {
	struct tcp_head	*next;
	u_int32_t	saddr;
	u_int32_t	smask;
	u_int32_t	daddr;
	u_int32_t	dmask;
	u_int16_t	sport_min;
	u_int16_t	sport_max;
	u_int16_t	dport_min;
	u_int16_t	dport_max;
	u_int8_t	flags;
	struct ip_rule	*rules;
	struct ip_rule	*rules_tail;
};

/* rule element */
struct relement {
	proc_match_match	m;
	void			*priv; /* cleanup with match->cleanup() */
	struct matcher		*match;
	u_int8_t		negate;
};

/* An IP rule */
struct ip_rule {
	struct ip_rule	*next;
	struct relement	*crit;
	struct relement *crit_end;
	struct alert	a;
};

/* Utility functions */
int iprule_addcrit(struct criteria *,
	struct ip_rule *,
	struct criteria *,
	unsigned int);
void iprule_cleanup(struct ip_rule *, unsigned int);

/* Shared between tcp and udp */
int tcp_udp_commit(struct tcp_head **,
	unsigned int *,
	unsigned int *,
	struct rule *,
	char **);

static inline int tcpip_traverse(
	struct generator *gen,
	struct packet *p, 
	unsigned int l, 
	struct ip_rule *r)
{
	register struct relement *c;
	
	while(r) {
		for(c=r->crit; c<r->crit_end; c++) {
			if ( !(c->m(p, c->priv, l, c->negate)) ) {
				goto failure;
			}
		}
		alert(gen, p, &r->a);
		return 1;
failure:
		r=r->next;
	}

	return 0;
}

#endif /* __PLUGIN_TCPIP__ */
