#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/unistd.h>
#include <string.h>
#include <plugin.h>
#include <cleanup.h>
#include <packet.h>
#include <signature.h>
#include <template.h>
#include <parser.h>

/* Snort rulefile parser
 * Genrally the parsing is done with turing state machines,
 * It isn't too hacky.
 *
 * TODO:
 * 	Full snort variable support
 * 	Support priority
 * 	Allow "config classification"
 */

/* Quick inline utility functions */
#define terminator(x) ( (x)=='\n' || (x)=='\r' )
static inline int whitespace(char x)
{
	char spc[]=" \t";
	char *p;
	for(p=spc; *p; p++)
		if ( *p==x ) return 1;
	return 0;
}

/* Maximum line length */
#define SNORT_MAXLEN 8192

/* Stuff we import */
proc_sig_rule signature_rule;
proc_sig_abort signature_abort;
proc_sig_criteria signature_criteria;
proc_sig_modifier signature_modifier;
proc_sig_alert signature_alert;
proc_sig_commit signature_commit;

/* Snort variables */
struct snort_var {
	struct snort_var	*next;
	char			*name;
	char			*value;
	int			negate;
};
struct snort_var *vars;

int snort_line(char *, int, char *);
int snort_var(char *, int, char *);

/* Snort keywords */
struct skw{
	char *name;
	int (*fn)(char *, int, char *);
};
struct skw snort_keywords[]={
	{"alert", snort_line},
	{"log", snort_line},
#if 0
	{"activate", NULL},
	{"dynamic", NULL},
	{"pass", NULL},
#endif
	{"var", snort_var},
	{NULL, NULL}
};

/* Snort parsing variables */
#define SNORT_PROTO 0
#define SNORT_SADDR 1
#define SNORT_SPORT 2
#define SNORT_DIR 3
#define SNORT_DADDR 4
#define SNORT_DPORT 5
#define SNORT_MAX 6
char *snort_head[SNORT_MAX];
int snort_negate[SNORT_MAX];
struct criteria *se_head;
struct criteria *se_tail;
char *snort_targetops;
char *snort_msg;

u_int32_t snort_sid, snort_rev;

/* These are modifiers */
struct criteria *snort_mod;
char *snort_modifiers[]={
	"nocase",
	"offset",
	"depth",
	"regex",

	/* Not implemented hacks */
	"reference",
	"classtype",
	"logto",
	NULL
};

int snort_commit_rule(int index);
int snort_iplist(int index);
struct {
	int var;
	int(*fn)(int);
}steps[]={
	{SNORT_SADDR, snort_iplist},
	{SNORT_DADDR, snort_iplist},
	{0, snort_commit_rule},
};

/* Error string */
#define ELEN 128
char estr[ELEN];

struct snort_var *snort_var_find(char *key)
{
	struct snort_var *ret;

	for(ret=vars; ret; ret=ret->next) {
		if ( !strcmp(key, ret->name) )
			return ret;
	}
	
	snprintf(estr, ELEN, "snort var not found: $%s", key);
	return NULL;
}

int snort_addcrit(char *key, char *val, int negate)
{
	int i;
	struct criteria *se;

	if ( !strcmp(key, "sid") ) {
		return strtouint(val, &snort_sid) ? 1 : 0;
	}
	if ( !strcmp(key, "rev") ) {
		return strtouint(val, &snort_rev) ? 1 : 0;
	}

	/* Modifier options */
	for(i=0; snort_modifiers[i]; i++) {
		if ( !strcmp(key, snort_modifiers[i]) ) {
			if ( !(se=malloc(sizeof(*se))) ) return 0;
			se->crit=key;
			se->args=val;
			se->negate=negate;
			se->next=snort_mod;
			snort_mod=se;
			return 1;
		}
	}

	if ( !strcmp(key, "msg") ) {
		snort_msg=val;
		return 1;
	}

	if ( !(se=malloc(sizeof(*se))) ) {
		return 0;
	}

	se->next=NULL;
	se->crit=key;
	se->args=val;
	se->negate=negate;

	if ( !se_head ) {
		se_head=se;
		se_tail=se;
	}else{
		se_tail->next=se;
		se_tail=se;
	}
	
	return 1;
}

/* Parse the body of a snort rule (the bit after the bracket) */
#define zero_body() key=NULL; val=NULL; neg=0; state=0;
int snort_parse_body(char *line)
{
	char *key=NULL;
	char *val=NULL;
	int neg=0;

	char *cur;
	int state=0;
	int inq=0;
	int esc=0;
	
	/* Snort state machine parser */
	for(cur=line; ; cur++) {
		if ( terminator(*cur) ) break;

		if ( state==0 ) { /* Anywhere */
			if ( !whitespace(*cur) ) {
				key=cur;
				state=1;
			}else if ( *cur==')' ) {
				*cur='\0';
				break;
			}
		}else if ( state==1 ) { /* Inside a key */
			if ( *cur==':' ) {
				*cur='\0';
				state=2;
			}else if ( *cur==';' && *(cur-1)!='\\' ) {
				*cur='\0';
				val=NULL;

				snort_addcrit(key, val, neg);
				zero_body();
			}
		}else if ( state==2 ) { /* Between key and value */
			if ( *cur=='!' ) {
				neg=1;
			}else if ( !whitespace(*cur) ) {
				if ( *cur=='\"' ) {
					inq=1;
				}

				val=cur + inq;
				state=3;
			}
		}else if ( state==3 ) { /* In a value */
			/* Enter quote marks */
			if ( !inq && *cur=='\"' ) {
				inq=1;
				continue;
			}

			/* Allow missing end semicolon */
			if ( !inq && *cur==')' ) {
				if ( *(cur-1)=='\"' ) *(cur-1)='\0';
				*cur='\0';
				
				snort_addcrit(key, val, neg);
				zero_body();
				continue;
			}

			/* Check for escape chars */
			if ( !esc && *cur=='\\' ) {
				esc=1;
				continue;
			}
			
			if ( inq && !esc && *cur=='\"' ) {
				/* End quotes */
				inq=0;
			}else if ( !inq && !esc && *cur==';' ) {
				/* End  */
				if ( *(cur-1)=='\"' ) *(cur-1)='\0';
				*cur='\0';

				snort_addcrit(key, val, neg);
				zero_body();
			}else{
				esc=0;
			}
		}
	}
	
	return 1;
}

/* Actually commit a rule into the signature 
 * subsystem */
int snort_commit_rule(int index)
{
	struct criteria *s;
	struct rule *r;

	/* Create a new rule */
	if ( !(r=signature_rule(snort_head[SNORT_PROTO])) ) {
		snprintf(estr, ELEN, "Unable to create new '%s' rule",
			snort_head[SNORT_PROTO]);
		return 0;
	}

	/* Alert stuff. Just use default alerting mechanism */
	if ( !snort_sid ) printf("snort: warning: rule has no sid: %s\n", snort_msg);
	if ( !signature_alert(r, snort_msg, snort_sid, snort_rev) ) {
		snprintf(estr, ELEN, "Unable to set rule target");
		goto abort;
	}

	/* Source and destination addresses */
	if ( !signature_criteria(r, snort_negate[SNORT_SADDR],
			"src", snort_head[SNORT_SADDR]) ){
		snprintf(estr, ELEN, "Unable add head criteria to rule");
		goto abort;
	}

	if ( !signature_criteria(r, snort_negate[SNORT_DADDR],
			"dst", snort_head[SNORT_DADDR]) ){
		snprintf(estr, ELEN, "Unable add head criteria to rule");
		goto abort;
	}

	
	/* Ports */
	if ( strcmp(snort_head[SNORT_SPORT], "any") ) {
		if ( !signature_criteria(r, snort_negate[SNORT_SPORT],
				"sport", snort_head[SNORT_SPORT]) ) {
			snprintf(estr, ELEN, "Unable add head criteria to rule");
			goto abort;
		}
	}
	if ( strcmp(snort_head[SNORT_DPORT], "any") ) {
		if ( !signature_criteria(r, snort_negate[SNORT_DPORT],
				"dport", snort_head[SNORT_DPORT]) ) {
			snprintf(estr, ELEN, "Unable add head criteria to rule");
			goto abort;
		}
	}

	/* All the modifiers */
	for(s=snort_mod; s; s=s->next) {
		if ( !signature_modifier(r, s->negate, s->crit, s->args) ) {
			goto abort2;
		}
	}
	
	/* And all the options... */
	for(s=se_head; s; s=s->next) {
		if ( !signature_criteria(r, s->negate, s->crit, s->args) ) {
			goto abort2;
		}
	}

	/* lets go */
	if ( !signature_commit(r) ) {
		snprintf(estr, ELEN, "Rule failed to commit: %s",
			snort_msg);
		goto abort;
	}
	
	return 1;
	
abort2:
	snprintf(estr, ELEN, "Unable add '%s' to rule", s->crit);
abort:
	signature_abort(r);
	return 0;
}

int snort_iplist(int index)
{
	char buf[128];
	char *cur, *ip, *tmp;
	size_t len=1;
	int ret=1;
	int i=steps[index++].var;

	if ( snort_head[i][0]!='[' ) {
		return steps[index].fn(index);
	}

	tmp=snort_head[i];
	snort_head[i]=buf;

	for(ip=cur=tmp+1; ; cur++, len++) {
		if ( *cur==',' || *cur==']' || *cur==0 ) {
			if ( len > sizeof(buf) ) {
				printf("snort: warning: ip in list too long!\n");
				break;
			}
			
			memcpy(buf, ip, len-1);
			buf[len-1]=0;

			if ( !steps[index].fn(index) ) {
				ret=0;
				break;
			}
			
			if ( *cur==',') {
				ip=cur+1;
				len=0;
				continue;
			}else break;
		}
	}

	snort_head[i]=tmp;
	return ret;
}

/* See if a string is a variable if so, swap it */
int snort_var_replace(char **ptr, int *n)
{
	struct snort_var *v;

	if ( **ptr=='$' ) {
		if ( !(v=snort_var_find(*ptr+1)) ) {
			return 0;
		}
		*ptr=v->value;
		*n=v->negate;
		return 1;
	}else return 1;
}

/* Parse a 'var' line */
int snort_var(char *fn, int ln, char *line)
{
	char *key=NULL, *val=NULL;
	char *cur;
	struct snort_var *vn;
	int end=0;
	int negate=0, nneg=0;
	int state=0;

	for(cur=line; ; cur++) {
		if ( terminator(*cur) ) {
			*cur=0;
			break;
		}

		if ( state==0 ) {
			if ( *cur=='!' && key ) {
				negate=1;
			}else if ( !whitespace(*cur) ) {
				if ( !key ) {
					key=cur;
					state=1;
				}else{
					if ( !val ) val=cur;
					continue;
				}
			}
		}else if ( state==1 ) {
			if ( whitespace(*cur) ) {
				*cur='\0';
				state=0;
				end=1;
			}
		}
	}
		
	if ( !end ) {
		snprintf(estr, ELEN, "Incomplete line");
		return 0;
	}

	/* TODO: This should be redundant */
	if ( !snort_var_replace(&val, &nneg) ) {
		negate = negate ^ nneg;
		return 0;
	}
	
	if ( (vn=snort_var_find(key)) ) {
		char *snew;

		if ( !(snew=strdup(val)) ) {
			goto err_nomem;
		}

		free(vn->value);
		vn->value=snew;
		vn->negate=negate;
	}else{
		if ( !(vn=malloc(sizeof(*vn))) ) {
			goto err_nomem;
			return 0;
		}

		if ( !(vn->name=strdup(key)) ) {
			free(vn);
			goto err_nomem;
		}

		if ( !(vn->value=strdup(val)) ) {
			free(vn->name);
			free(vn);
			goto err_nomem;
		}

		vn->negate=negate;
		vn->next=vars;
		vars=vn;
	}

	return 1;
 err_nomem:
	snprintf(estr, ELEN, "Out of memory for var '%s'", key);
	return 0;
}

/* Snort up each line of the config  */
int snort_line(char *fn, int ln, char *line)
{
	struct criteria *s;
	char *cur;
	int i;
	int end=0;
	int state=0;
	char any_sip[]="0/0";
	char any_dip[]="0/0";

	/* Zero all global values */
	estr[0]=0;
	snort_sid=0;
	snort_rev=0;
	for(i=0; i<SNORT_MAX; i++) {
		snort_negate[i]=0;
	}

	/* State machine parser, parse out first 7 elements
	 * seperated by any amount of whitespace ending in a '(' */
	for(i=0, cur=line; ; cur++) {
		if ( terminator(*cur) ){
			break;
		}

		if ( state == 0 ) {
			if ( *cur=='(' ) {
				if ( i==SNORT_MAX ) end=1;
				break;
			}
			
			if ( *cur=='!' ) {
				snort_negate[i]=1;
			}else if ( !whitespace(*cur) && *cur ) {
				if ( i >= SNORT_MAX ) break;

				snort_head[i]=cur;
				state=1;
				i++;
			}
		}else if ( state == 1 ) {
			if ( whitespace(*cur) ) {
				*cur='\0';
				state=0;
			}
		}
	}

	/* Parse error */
	if ( !end ) {
		snprintf(estr, ELEN, "Incomplete line");
		return 0;
	}

	/* Its either "<-", "->" or "<>" */ 
	if ( strlen(snort_head[SNORT_DIR]) !=2 ) {
		snprintf(estr, ELEN, "Invalid direction '%s'",
			snort_head[SNORT_DIR]);
		return 0;
	}

	/* Parse the rest of the rule */
	se_head=NULL;
	se_tail=NULL;
	snort_mod=NULL;
	snort_msg=NULL;
	snort_targetops=NULL;
	snort_parse_body(++cur);

	/* Do variable substitutions */
	if ( !snort_var_replace(&snort_head[SNORT_SADDR],
		&snort_negate[SNORT_SADDR]) ) return 0;
	if ( !snort_var_replace(&snort_head[SNORT_SPORT],
		&snort_negate[SNORT_SPORT]) ) return 0;
	if ( !snort_var_replace(&snort_head[SNORT_DADDR],
		&snort_negate[SNORT_DADDR]) ) return 0;
	if ( !snort_var_replace(&snort_head[SNORT_DPORT],
		&snort_negate[SNORT_DPORT]) ) return 0;

	/* any == "0/0" for IP addresses */
	if ( !strcmp(snort_head[SNORT_SADDR], "any") ) {
		snort_head[SNORT_SADDR]=any_sip;
	}

	if ( !strcmp(snort_head[SNORT_DADDR], "any") ){
		snort_head[SNORT_DADDR]=any_dip;
	}

	/* Go forwards */
	if ( snort_head[SNORT_DIR][1]=='>' ) {
		if ( !steps[0].fn(0) ) {
			return 0;
		}
	}

	/* Go backwards */
	if ( snort_head[SNORT_DIR][0]=='<' ) {
		char *tadr, *tprt;
		int tna, tnp;

		/* Swap IPs and ports */
		tadr=snort_head[SNORT_SADDR];
		snort_head[SNORT_SADDR]=snort_head[SNORT_DADDR];
		snort_head[SNORT_DADDR]=tadr;

		tna=snort_negate[SNORT_SADDR];
		snort_negate[SNORT_SADDR]=snort_negate[SNORT_DADDR];
		snort_negate[SNORT_SPORT]=tna;

		tprt=snort_head[SNORT_SPORT];
		snort_head[SNORT_SPORT]=snort_head[SNORT_DPORT];
		snort_head[SNORT_DPORT]=tprt;
		
		tnp=snort_negate[SNORT_SPORT];
		snort_negate[SNORT_DADDR]=tna;
		snort_negate[SNORT_SPORT]=snort_negate[SNORT_DPORT];

		if ( !steps[0].fn(0) ) {
			return 0;
		}
	}

	for(s=se_head; s; ) {
		struct criteria *tmp=s;
		s=s->next;
		free(tmp);	
	}
	
	for(s=snort_mod; s; ) {
		struct criteria *tmp=s;
		s=s->next;
		free(tmp);	
	}
	
	return 1;
}

/* Look at the first word of a snort line and decide
 * what kind of line it is and how to deal with it
*/
int snort_dispatch(char *fn, int ln, char *line)
{
	struct skw *k;
	char *cur, *kw=NULL;
	int state=0;
	int end=0;

	for(cur=line; ;cur++) {
		if ( terminator(*cur) ) {
			break;
		}

		if ( state==0 ) {
			if ( !whitespace(*cur) ) {
				if ( !kw ) {
					kw=cur;
					state=1;
				}else{
					break;
				}
			}
		}else if ( state==1 ) {
			if ( whitespace(*cur) ) {
				*cur=0;
				state=0;
				end=1;
			}
		}
	}

	if ( !end ) return 0;

	for(k=snort_keywords; k->name; k++) {
		if ( !strcmp(kw, k->name) )
			return k->fn(fn, ln, cur);
	}

	snprintf(estr, ELEN, "Unsupported snort keyword: '%s'", kw);
	return 0;
}

/* Parse the snort file, line by line */
int snort_run(char *fn)
{
	FILE *f;
	static char buf[SNORT_MAXLEN];
	int line=0;

	if ( !fn ) return 0;

	if ( (f=fopen(fn, "r"))==NULL ) {
		printf("snort: error: unable to open file '%s'\n", fn);
		return 0;
	}

	/* iterate each line in the file */
	while(fgets(buf, SNORT_MAXLEN, f)) {
		char *p=buf;
		
		line++;

		while ( *p && whitespace(*p) ) p++;
		if ( terminator(*p) || *p=='#' ) continue;
		
		if ( !snort_dispatch(fn, line, p) ) {
			printf("snort: %s:%i : %s\n", fn, line, estr);
			fclose(f);
			return 0;
		}
	}

	fclose(f);
	return 1;
}

int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	proc_parser_add parser_add;
	int ok=0;
	
	plugin_check(in, out);
	
	PLUGIN_ID("parser.snort", "Snort ruleset files");
	PLUGIN_VERSION(2, 0);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@scaramanga.co.uk");
	PLUGIN_LICENSE("GPL");
	
	if ( !(parser_add=in->import("parser.add")) )
		return PLUGIN_ERR_OBJECT;
	
	if ( !(signature_rule=in->import("signature.rule")) )
		return PLUGIN_ERR_OBJECT;
	
	if ( !(signature_abort=in->import("signature.abort")) )
		return PLUGIN_ERR_OBJECT;
	
	if ( !(signature_criteria=in->import("signature.criteria")) )
		return PLUGIN_ERR_OBJECT;

	if ( !(signature_modifier=in->import("signature.modifier")) )
		return PLUGIN_ERR_OBJECT;
	
	if ( !(signature_alert=in->import("signature.alert")) )
		return PLUGIN_ERR_OBJECT;

	if ( !(signature_commit=in->import("signature.commit")) )
		return PLUGIN_ERR_OBJECT;

	ok+=parser_add("snort", snort_run);
	
	return (ok) ? PLUGIN_ERR_OK : PLUGIN_ERR_FAIL;
}

int _PLUGIN_UNLOAD (int code) {
	return PLUGIN_ERR_OK;
}
