/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002 Gianni Tedesco
 * This program is released under the terms of the GNU GPL version 2
 *
 * 	This plugin implements the (experimental) firestorm transport
 * 	for alert information. The data is saved to a file currently
 * 	but the serialised data could also be sent across local IPC
 * 	or network connections. This format will form the basis of
 * 	all future alerting performed by firestorm.
 *
 * TODO:
 * 	o Hack decode plugins to implement serialisation routines
 * 	o Move code in to firestorm core, sort out alert modules
 * 	o Implement zlib packet compression (optional)
 * 	o Investigate portability requirements
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <packet.h>
#include <cleanup.h>
#include <capture.h>
#include <capdev.h>
#include <signature.h>
#include <decode.h>
#include <alert.h>
#include <target.h>
#include <plugin.h>
#include <template.h>
#include <args.h>
#include <time.h>
#include <elog.h>

#include <sys/uio.h>
#include <netinet/in.h>

/* Imported functions */
proc_args_parse args_parse;

/* Private data structure */
struct elog_priv {
	char *fn;
	int fd;
	char *buf;
	size_t buflen;
};

void elog_a(struct generator *gen,
	struct packet *pkt,
	struct alert *a,
	void *priv)
{
	static char decode_buf[4096];
	static char *ptr_end=decode_buf+sizeof(decode_buf);
	struct elog_dhdr *dh;
	char *ptr=decode_buf;
	struct elog_priv *p=priv;
	struct elog_pkthdr ph;
	int gen_len, alert_len;
	u_int32_t tot_len;
	struct iovec iov[6];
	char padding[4]={0,0,0,0};
	int i=0;
	int j;

	if ( !pkt ) return;

	/* Calculate length of generator strings */
	gen_len=strlen(gen->name)+1;
	alert_len=strlen(a->alert)+1;

	tot_len = sizeof(ph) + pkt->caplen +
		gen_len + alert_len;

	/* Construct header */
	ph.reclen=tot_len;
	ph.pflags=pkt->flags;
	ph.sid=a->sid;
	ph.rev=a->rev;
	ph.gen_len=gen_len;
	ph.alert_len=alert_len;
	ph.pkt_len=pkt->len;
	ph.pkt_caplen=pkt->caplen;
	ph.decode_len=0;
	ph.ts.tv_sec=pkt->time.tv_sec;
	ph.ts.tv_usec=pkt->time.tv_usec;
	iov[i].iov_base=&ph;
	iov[i++].iov_len=sizeof(ph);

	/* Construct the decode data */
	for(j=0; j<pkt->llen; j++) {
		int nlen, real;
		
		if ( !pkt->layer[j].proto ) break;

		real=nlen=strlen(pkt->layer[j].proto->name)+1;

		/* Round to the nearest multiple of 4 */
		if ( nlen & 0x3 ) nlen+=(4-(nlen&0x3));

		if ( ptr+nlen+sizeof(*dh) > ptr_end ) {
			printf("elog: decode buffer too small!\n");
			return;
		}

		/* Fill in the header */
		dh=(struct elog_dhdr *)ptr;
		ptr+=sizeof(*dh);
		dh->tot_len=(sizeof(*dh)+nlen)>>2;
		dh->name_len=nlen>>2;
		dh->pkt_ofs=htons(pkt->layer[j].h.raw - pkt->base);
		dh->flags=htonl(pkt->layer[j].flags);

		/* Then the protocol name  */
		memcpy(ptr, pkt->layer[j].proto->name, real);
		ptr+=nlen;
	}

	/* write decode data */
	iov[i].iov_base=decode_buf;
	iov[i].iov_len=ptr-decode_buf;
	ph.decode_len=(iov[i].iov_len>>2);
	ph.reclen+=iov[i++].iov_len;

	/* Write out packet data */
	iov[i].iov_base=pkt->base;
	iov[i++].iov_len=pkt->caplen;
	
	/* Write out generator and alert strings */
	iov[i].iov_base=gen->name;
	iov[i++].iov_len=gen_len;
	iov[i].iov_base=a->alert;
	iov[i++].iov_len=alert_len;

	/* Pad to multiples of 4 bytes */
	if ( ph.reclen & 3 ) {
		iov[i].iov_base=padding;
		iov[i].iov_len=(4-(ph.reclen&3));
		ph.reclen+=iov[i++].iov_len;
	}

	/* Convert to network byte order  */
	ph.reclen=htonl(ph.reclen);
	ph.sid=htonl(ph.sid);
	ph.rev=htonl(ph.rev);
	ph.pflags=htonl(ph.pflags);
	ph.decode_len=htons(ph.decode_len);
	ph.pkt_len=htonl(ph.pkt_len);
	ph.pkt_caplen=htonl(ph.pkt_caplen);
	ph.ts.tv_sec=htonl(ph.ts.tv_sec);
	ph.ts.tv_usec=htonl(ph.ts.tv_usec);

	/* Actually write it all */
	if ( writev(p->fd, iov, i)<0 ) {
		printf("elog: %s: writev(): %s", p->fn, get_err());
	}

	/* zero out the padded bits */
	memset(decode_buf, 0, ntohs(ph.decode_len));
}

/* Configuration callbacks */
int cb_file(struct arg *a, void *priv)
{
	struct elog_priv *p=priv;
	return (p->fn=strdup(a->val.v_str)) ? 1 : 0;
}

int elog_openfile(struct elog_priv *p)
{
	struct efile_hdr fhdr;
	int ret;
	
	if ( (p->fd=open(p->fn, O_WRONLY|O_TRUNC|O_CREAT, 0640))<0 ) {
		printf("elog: %s: open(): %s\n", p->fn, get_err());
		return 0;
	}
	
	/* Construct header */
	fhdr.magic=htonl(EF_MAGIC);
	fhdr.flags=0;
	fhdr.vers_minor=EF_VERS_MIN;
	fhdr.vers_major=EF_VERS_MAJ;
	
	/* Write header */
	ret=write(p->fd, &fhdr, sizeof(fhdr));
	if ( ret<sizeof(fhdr) ) {
		if ( ret<0 ) {
			printf("elog: %s: write(): %s\n", 
				p->fn, get_err());
		}else{
			printf("elog: %s: writing file header: truncated\n",
				p->fn);
		}
		close(p->fd);
		free(p->fn);
		free(p);
		return 0;
	}
		
	return 1;
}


int elog_v(char *args, void **ptr)
{
	struct elog_priv *priv;
	struct arg elog_args[]={
		{"file", ARGTYPE_STRING, cb_file},
		{NULL, ARGTYPE_NOP, NULL},
	};

	if ( !(priv=calloc(1, sizeof(*priv))) )
		return 0;

	priv->buflen=1024;

	if ( args ) {
		switch ( args_parse(elog_args, args, priv) ) {
		case -1:
			printf("log: parse error: %s\n", args);
		case 0: /* fall through */
			free(priv);
			return 0;
		default:
			break;
		}
	}

	if ( !priv->fn ) {
		printf("elog: must specify a filename\n");
		free(priv);
		return 0;
	}

	if ( !(priv->buf=malloc(priv->buflen)) ) {
		printf("elog: malloc(): %s\n", get_err());
		free(priv->fn);
		free(priv);	
		return 0;
	}

	if ( !elog_openfile(priv) ) {
		free(priv->fn);
		free(priv);
		return 0;
	}
	
	*ptr=priv;
	return 1;
}

void elog_h(void *priv)
{
	struct elog_priv *p=priv;

	if ( !p ) return;
	
	if ( p->fd>2 ) {
		if ( fsync(p->fd) )
			printf("elog: fsync(): %s\n", get_err());
		if ( close(p->fd) )
			printf("elog: close(): %s\n", get_err());
		if ( !elog_openfile(p) ) {
			printf("elog: error reopening logfile\n");
			return;
		}
	}
}

void elog_c(void *priv)
{
	struct elog_priv *p=priv;
	
	if ( !p )  return;

	/* Don't close if we are stdin/stdout */
	if ( p->fd>2 ) {
		if ( fsync(p->fd) )
			printf("elog: fsync(): %s\n", get_err());
		if ( close(p->fd) ) 
			printf("elog: close(): %s\n", get_err());
	}

	if ( p->fn ) free(p->fn);

	free(p);
}

int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	proc_target_add target_add;
	static struct target alert_t[]={
		target_init("elog", elog_a, elog_v, elog_h, elog_c),
		target_null()
	};
	
	plugin_check(in, out);
	
	PLUGIN_ID("target.elog", "Extended log output");
	PLUGIN_VERSION(0,1);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@scaramanga.co.uk");
	PLUGIN_LICENSE("GPL");
	
	if ( !(args_parse=in->import("args.parse")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(target_add=in->import("target.add")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	return target_add(alert_t) ?
		PLUGIN_ERR_OK : PLUGIN_ERR_FAIL;
}

int _PLUGIN_UNLOAD (int code) {
	return PLUGIN_ERR_OK;
}
